<cfscript>
import algorithms.sorting.SortAlgorithms;

/**
 * Component of merging algorithms.
 * 
 * @output false
 */
component
{	
	/**
	 * Constructor
	 *
	 * @output false
	 */
	public any function init()
	{
		SortAlgorithms = new SortAlgorithms();
		
		return this;
	}
	
	/**
	 * Combines two arrays of sorted values using a sort.
	 *
	 * @a    Array 1
	 * @b    Array 2
	 * @type Sort type. Same choices as array sort.
	 * 
	 * @output false
	 */	
	public array function mergeArraysSort(required array a, required array b, required string type)
	{
		a.addAll(b);
		arraySort(a,type);
		return a;
	}
	
	/**
	 * Combines two array of structs by computing a sorting key.
	 *
	 * @a    Array 1
	 * @b    Array 2
	 * @paths An array of paths to simple values to sort by
	 * @type Type of sort to do, choices same as arraySort().
	 *
	 * @output false
	 */
	public array function mergeArrayKeySort(required array a, required array b, required array paths, string type="text")
	{
		var result = [];
		var keyMap = {};
		
		a.addAll(b);
		
		for( var item in a )
		{
			var computed = [];
						
			for( var path in paths )
			{	
				arrayAppend(computed,findOneOf(".(",path) ? evaluate("item.#path#") : item[path]);
			}
			
			computed = arrayToList(computed);
			
			if( not structKeyExists(keyMap,computed) )
				keyMap[computed] = [];
			
			arrayAppend(keyMap[computed],item);
		}
		
		var keys = structKeyArray(keyMap);
		arraySort(keys,type);
		
		for( var key in keys )
			result.addAll(keyMap[key]);
			
		return result;
	}
	
	/**
	 * OPTIMIZED.
	 * 
	 * Combines two array of structs by computing a sorting key.
	 *
	 * @a    Array 1
	 * @b    Array 2
	 * @paths An array of paths to simple values to sort by
	 * @type Type of sort to do, choices same as arraySort().
	 *
	 * @output false
	 */
	public array function mergeArrayKeySortOptimized(required array a, required array b, required array paths, string type="text")
	{
		var result = [];
		var keys = [];
		
		a.addAll(b);
		
		for( var item in a )
		{
			var computed = [];
						
			for( var path in paths )
			{	
				arrayAppend(computed,findOneOf(".(",path) ? evaluate("item.#path#") : item[path]);
			}
			
			arrayAppend(computed,arrayLen(keys)+1);
			arrayAppend(keys,arrayToList(computed));
		}
		
		arraySort(keys,type);
		
		for( var key in keys )
			arrayAppend(result,a[listLast(key)]);
			
		return result;
	}
	
	/**
	 * Combines two array of structs by computing a sorting key and uses a query sort.
	 *
	 * @a    Array 1
	 * @b    Array 2
	 * @paths An array of paths to simple values to sort by
	 * @type Type of sort to do, choices same as arraySort().
	 *
	 * @output false
	 */
	public array function mergeArrayKeySortQuery(required array a, required array b, required array paths)
	{		
		a.addAll(b);
		
		return SortAlgorithms.querySortEnhanced(a,paths);
	}
		
	/**
	 * Combines two arrays of sorted values using a predicate.
	 *
	 * @a          Array 1
	 * @b          Array 2
	 * @comparator Function to sort with. Must take two arguments (a,b) to compare.
	 * 
	 * @output false
	 */
	public array function mergeArrays(required array a, required array b, required any comparator)
	{
		var result = [];
		var i=1;
		var j=1;
		
		while( i <= arrayLen(a) )
		{			
			if( j > arrayLen(b) or comparator(a[i],b[j]) <= 0 )
			{
				arrayAppend(result,a[i]);
				++i;
			}
			else
			{
				arrayAppend(result,b[j]);
				++j;
			}
		}
		
		for(; j <= arrayLen(b); ++j )
			arrayAppend(result,b[j]);
			
		return result;
	}
}
</cfscript>