<cfscript>
import algorithms.Utils;

/**
 * Component of sorting algorithms.
 * 
 * @output false
 */
component
{
	
	/**
	 * Constructor
	 *
	 * @output false
	 */
	public any function init()
	{
		Utils = new Utils();
		
		return this;
	}
	
	/**
	 * Sorts an array of structs by computing a sorting key.
	 *
	 * @list The array to sort
	 * @paths An array of paths to simple values to sort by
	 * @type Type of sort to do, choices same as arraySort().
	 *
	 * @output false
	 */
	public array function keySort(required array list, required array paths, string type="text")
	{
		var result = [];
		var keys = [];
		
		for( var item in list )
		{
			var computed = [];
						
			for( var path in paths )
			{	
				arrayAppend(computed,findOneOf(".(",path) ? evaluate("item.#path#") : item[path]);
			}
			
			arrayAppend(computed,arrayLen(keys)+1);
			arrayAppend(keys,arrayToList(computed));
		}
		
		arraySort(keys,type);
		
		for( var key in keys )
			arrayAppend(result,list[listLast(key)]);
			
		return result;
	}
	
	/**
	 * Sorts an array of structs by conversion to a struct.
	 *
	 * @list The array to sort
	 * @type The type of sort to do. Choices are the same as structSort()
	 * @path Path to element in struct to sort by.
	 *
	 * @output false
	 */
	public array function structArraySort(required array list, required string type, required string path)
	{
		var map = {};
		var result = [];
		
		for( var i=1; i <= arrayLen(list); ++i )
			map[i] = list[i];
			
		for( var i in structSort(map,type,"asc",path) )
			arrayAppend(result,list[i]);
		
		return result;
	}
		
	/**
	 * Sorts an array of structs by conversion to a query. Columns to sort by must be simple
	 * values, or paths to them.
	 *
	 * @list    The array to sort
	 * @orderBy Order by clause of properties to sort by
	 *
	 * @output false
	 */
	public array function querySort(required array list, required array orderBy)
	{
		var query = queryNew("x_element_index");
		var columns = [];
		var result = [];
		
		if( not arrayLen(list) )
			return [];
			
		for( var column in orderBy )
		{
			column = trim(listFirst(column," "));
			arrayAppend(columns,column);
			queryAddColumn(query,column,[]);
		}
		
		for( var i=1; i <= arrayLen(list); ++i )
		{
			queryAddRow(query);
			
			for( var column in columns )
			{
				querySetCell(query,column,list[i][column]);
			}
			
			querySetCell(query,"x_element_index",i);
		}
		
		/*
		query = new Query(
			dbtype="query",
			sql="select x_element_index as index from queryValue order by #arrayToList(orderBy)#",
			queryValue=query
		)
			.execute()
			.getResult();
		*/
		// Using the new Query.cfc is super slow, faster to write your own wrapper (5ms vs 25ms)
		query = Utils.executeQueryOfQueries(
			"select x_element_index as index from sourceQuery order by #arrayToList(orderBy)#",
			{sourceQuery=query}
		);
			
		for( var i=1; i <= query.recordCount; i++ )
			arrayAppend(result,list[query.index[i]]);
			
		return result;
	}
	
	/**
	 * Sorts an array of structs or objects by conversion to a query. 
	 * Columns to sort by must be simple values, or paths to them.
	 *
	 * ex. ["getPackage().getParentEvent().getName() asc","getLastName() desc"]
	 *
	 * @list    The array to sort
	 * @orderBy Order by clause of properties to sort by
	 *
	 * @output false
	 */
	public array function querySortEnhanced(required array list, required array orderBy)
	{
		var query = queryNew("item");
		var sqlOrderBy = [];
		var result = [];
		
		if( not arrayLen(list) )
			return [];
			
		for( var part in orderBy )
		{
			var path = trim(listFirst(part," "));
			var name = "x_" & hash(path);
			var values = [];
			
			for( var item in list )
				arrayAppend(values,findOneOf(".(",path) ? evaluate("item.#path#") : item[path]);
			
			arrayAppend(sqlOrderBy,name & " " & (find(" ",part) ? listLast(part," ") : ""));
			queryAddColumn(query,name,values);
		}
		
		for( var i=1; i <= arrayLen(list); ++i )
			querySetCell(query,"item",list[i],i);
		
		query = Utils.executeQueryOfQueries(
			"select item from sourceQuery order by #arrayToList(sqlOrderBy)#",
			{sourceQuery=query}
		);
	
		for( var i=1; i <= arrayLen(list); ++i )
			arrayAppend(result,query.item[i]);
	
		
		return result;
	}
		
	/**
	 * Sorts an array of values using a comparator function with 
	 * the QuickSort algorithm.
	 * 
	 * @list       The array to sort
	 * @comparator Function to sort with. Must take two arguments (a,b) to compare.
	 * @desc       Sort in descending Order
	 *
	 * @output false
	 */
	public array function quickSort(required array list, required any comparator, boolean desc=false)
	{
		if( arrayLen(list) <= 1 )
			return list;
		
		var less = [];
		var pivot = [list[1]];
		var greater = [];
		
		for( var i=2; i <= arrayLen(list); ++i )
		{
			var result = (desc)
				? comparator(pivot[1],list[i]) 
				: comparator(list[i],pivot[1]);
			
			if( result < 0 )
			{
				arrayAppend(less,list[i]);
			}
			else if( result > 0 )
			{
				arrayAppend(greater,list[i]);
			}
			else
			{
				arrayAppend(pivot,list[i]);
			}
		}
		
		less = quickSort(less,comparator,desc);
		less.addAll(pivot);
		less.addAll(quickSort(greater,comparator,desc));
		
		return less;
	}
}
</cfscript>